/*
 * MqttClient.cpp
 *
 *  Created on: 08.05.2015
 *      Author: wn
 */

#include "MqttClient.h"

#include <PubSubClient.h>
#include <PString.h>
#include <Streaming.h>
#include <Metro.h>
#include <WString.h>
#include <Print.h>
#include "cmd.h"
#include "config.h"
#include "reset.h"


char MQTT_BROKER_DEFAULT[] = "192.168.75.1";
const uint16_t MQTT_PORT = 1883;




String MqttConfig::exec(String params) {
  String res = "failed";

  Print *out = m_server;
  MqttClient *mc = (MqttClient*)m_mqttClient;
  if (params.equalsIgnoreCase("help")) {
    *out << "help ..... this help page" << endl;
    *out << "show ..... shows the whole configuration"  << endl;
    *out << "broker ... set the broker's address" << endl;
    *out << "           followed by a restart, you need to" << endl;
    *out << "           about a minute" << endl;
    *out << "add ...... add an mbus client, params: token address period" << endl;
    *out << "del ...... delete an mbus client, params: index" << endl;
    *out << "reset .... reset configuration" << endl;
    *out << "           followed by a restart, you need to" << endl;
    *out << "           about a minute" << endl;
    *out << endl;
    res = "done";
  } else if (params.equalsIgnoreCase("show")) {
    // show the whole configuration
    *out << "Clients:" << endl;
    *out << "(index, token, address, period)" << endl;
    for (uint8_t i = 0; i < NUM_OF_DEVICES; i++) {
      *out << i << ": " << mc->m_mbusDevTuple[i].token << ", " <<
          mc->m_mbusDevTuple[i].address << ", " <<
          mc->m_mbusDevTuple[i].queryPeriod << endl;
    }
    *out << "Broker: " << mc->m_mqttBroker << endl;
    *out << "Watchdog resets: " << WDOG_RSTCNT << endl;
    res = "done";
  } else if (params.startsWith("broker ")) {
    // set the broker's address
    int space = params.indexOf(' ');
    if (space == -1) {
      res = "missing argument";
    } else {
      String broker = params.substring(space+1);
      strcpy(mc->m_mqttBroker, broker.c_str());
      configWrite(CONFIG_BROKER, sizeof(mc->m_mqttBroker), (char*)mc->m_mqttBroker);
      *out << "Stopping gateway, wait for reset" << endl;
      while (true);
      res = "done";
    }
  } else if (params.startsWith("add ")) {
    // add an mbus client, params: token address period
    char paramBuf[64];
    char *paramPtr;
    strncpy(paramBuf, params.c_str(), sizeof(paramBuf));
    paramPtr = paramBuf;
    char *tokenStr = 0;
    char *addressStr = 0;
    char *periodStr = 0;
    if ((paramPtr != 0) && (*paramPtr != 0)){
      // command
      strsep(&paramPtr, " ");
    }
    if ((paramPtr != 0) && (*paramPtr != 0)){
      tokenStr = strsep(&paramPtr, " ");
    }
    if ((paramPtr != 0) && (*paramPtr != 0)){
      addressStr = strsep(&paramPtr, " ");
    }
    if ((paramPtr != 0) && (*paramPtr != 0)){
      periodStr = strsep(&paramPtr, " ");
    }

    if ((tokenStr != 0) && (*tokenStr != 0) &&
        (addressStr != 0) && (*addressStr != 0) &&
        (periodStr != 0) && (*periodStr != 0)) {
      int token = atoi(tokenStr);
      bool validToken = ! ((token <= 0) || (token > 250));
      int address = atoi(addressStr);
      bool validAddress = ! ((address <= 0) || (address > 250));
      int period = atoi(periodStr);
      bool validPeriod = ! ((period < 10) || (period > 3600));

      if (validToken && validAddress && validPeriod) {
        uint8_t i = 0;
        while (mc->m_mbusDevTuple[i].address != 0) {
          i++;
        }
        mc->m_mbusDevTuple[i] = { (uint8_t)token, (uint8_t)address, (uint16_t)period };
        configWrite(CONFIG_DEVICES, sizeof(mc->m_mbusDevTuple), (char*)mc->m_mbusDevTuple);
        res = "done";
      } else {
        if (! validToken) {
          *out << "Invalid token " << tokenStr << endl;
        }
        if (! validAddress) {
          *out << "Invalid address " << addressStr << endl;
        }
        if (! validPeriod) {
          *out << "Invalid period " << periodStr << endl;
        }
        res = "failure";
      }
    } else {
      *out << "not enough arguments" << endl;
      res = "failure";
    }


  } else if (params.startsWith("del ")) {
    // delete an mbus client, params: index
    int space = params.indexOf(' ');
    if (space == -1) {
      res = "missing argument";
    } else {
      String arg0 = params.substring(space);
      int idx = atoi(arg0.c_str());
      if (idx < 0 || idx >= NUM_OF_DEVICES) {
        res = "illegal index";
      } else {
        *out << "Index " << arg0 << " (" << idx << ") to be deleted" << endl;
        mc->m_mbusDevTuple[idx] = {0, 0, 0, 0};
        configWrite(CONFIG_DEVICES, sizeof(mc->m_mbusDevTuple), (char*)mc->m_mbusDevTuple);
        res = "done";
      }
    }
  } else if (params.equalsIgnoreCase("reset")) {
    configReset();
    *out << "Stopping gateway, wait for reset" << endl;
    resetDevice();
    res = "done";
  }
  return res;
}



void callback(char* topic, byte* payload, unsigned int length) {
}


Metro secondTick = Metro(1000);


MqttClient::MqttClient(RequestSender *meterBusMaster) : m_mqttConfig(this),
                m_client(), m_meterBusMaster(meterBusMaster), m_mqttClient(m_client),
                m_disconnectState(3), m_disconnectTime(millis()), m_uptime(0), m_deviceIdx(0)
{
}

void MqttClient::sendResponse(uint8_t *responseBuffer, uint16_t responseBufferLength, uint8_t token) {
  char strbuf[1024];
  memset(strbuf, sizeof(strbuf), 0);
  PString buf = PString(strbuf, sizeof(strbuf));


  buf << "{ \"metadata\": { \"device\": \"MeterbusHub\", " <<
      "\"token\": " << token <<
      "}, " <<
      "\"data\": {" <<
      "\"uptime\": " << m_uptime << ", " <<
      "\"telegram\": \"";

  uint16_t i = 0;
  while (true) {
    if (responseBuffer[i] <= 0x0f) {
      buf.print("0");
    }
    buf.print(responseBuffer[i], HEX);
    buf.print(" ");
    i++;
    if (i == responseBufferLength) {
      break;
    }
  }

  buf << "\"}}";

  if (m_disconnectState == 0) {
    //Serial << "publishing " << strbuf << endl;
    //Serial << "length: " << buf.length() << endl;
    m_mqttClient.publish("IoT/Measurement/MeterbusHub", strbuf);
  } else {
    Serial << "no MQTT connection, message lost: " << endl <<
        strbuf << endl;
  }
}

void MqttClient::sendError(uint8_t code, uint8_t token) {
  String msg = String("{ \"metadata\": { \"device\": \"MeterbusHub\", \"error\": ")
      + code + String(", \"token\": ") + token + String(" }, \"data\": { \"uptime\": ") + m_uptime + String("}}");
  if (m_disconnectState == 0) {
    //Serial << "publishing " << msg << endl;
    //Serial << "length: " << msg.length() << endl;
    m_mqttClient.publish("IoT/Failure/MeterbusHub", (char*)msg.c_str());
  } else {
    Serial << "no MQTT connection, message lost: " << msg << endl;
  }
}

void MqttClient::begin(CmdServer *cmdServer) {
  m_mqttConfig.registerYourself(cmdServer);

  if (! configIsValid()) {
    for (uint8_t i = 0; i < NUM_OF_DEVICES; i++) {
      m_mbusDevTuple[i] = { 0, 0, 0, 0 };
    }
    m_mbusDevTuple[0] = { 1, 0x53, 10, 0 }; // light meter
    m_mbusDevTuple[1] = { 2, 32,   10, 0 }; // electrity
    configWrite(CONFIG_DEVICES, sizeof(m_mbusDevTuple), (char*)m_mbusDevTuple);

    strcpy(m_mqttBroker, MQTT_BROKER_DEFAULT);
    configWrite(CONFIG_BROKER, sizeof(m_mqttBroker), (char*)m_mqttBroker);
  }


  configRead(CONFIG_DEVICES, sizeof(m_mbusDevTuple), (char*)m_mbusDevTuple);
  configRead(CONFIG_BROKER, sizeof(m_mqttBroker), (char*)m_mqttBroker);



  m_mqttClient = PubSubClient(m_mqttBroker, MQTT_PORT, callback, m_client);
}

void MqttClient::exec() {
  //Serial << "*** a" << endl;
  if ((m_disconnectState == 0) && (! m_mqttClient.loop())) {
    m_disconnectState = 1;
  }

  //Serial << "*** b, " << m_disconnectState << endl;
  switch (m_disconnectState) {
  case 0:
    // Serial.println("discState 0");
    // everything fine
    break;
  case 1:
    Serial.println("MQTT Connection lost");
    m_mqttClient.disconnect();
    m_disconnectTime = millis();
    m_disconnectState = 2;
    break;
  case 2:
    if (m_disconnectTime + 2000 < millis()) {
      m_disconnectState = 3;
    }
    break;
  case 3:
    Serial << "Trying to re-connect" << endl;
    if (m_mqttClient.connect("MeterbusHub")) {
      Serial << "MQTT connected" << endl;
      m_disconnectTime = millis();
      m_disconnectState = 0;
    } else {
      Serial << "no success" << endl;
      m_disconnectState = 1;
    }
    break;
  default:
    m_disconnectState = 0;
    break;
  }

  //Serial << "*** c" << endl;
  if (secondTick.check() == 1) {
    m_uptime++;
    //Serial << "Tick " << m_uptime << endl;

    byte wdogCnt = WDOG_RSTCNT;
    String msg = String("{ \"metadata\": { \"device\": \"MeterbusHub\" }, \"data\": { \"uptime\": ") + m_uptime + String(", \"watchdogCnt\": ") + wdogCnt + String("}}");
    if (m_disconnectState == 0) {
      m_mqttClient.publish("IoT/Heartbeat/MeterbusHub", (char*)msg.c_str());
    } else {
      Serial << "no MQTT connection, message lost: " << msg << endl;
    }

    for (uint8_t i = 0; i < NUM_OF_DEVICES; i++) {
      if ((m_mbusDevTuple[i].address != 0) && (m_mbusDevTuple[i].timer != 0)) {
        m_mbusDevTuple[i].timer -= 1;
//        if (m_mbusDevTuple[i].timer == 0) {
//          Serial << "Device " << m_mbusDevTuple[i].token << " ready for request" << endl;
//        } else {
//          Serial << "Device " << m_mbusDevTuple[i].token << " not ready for request, timer: " << m_mbusDevTuple[i].timer << endl;
//        }
      }
    }

    //Serial << "while in" << endl;
    while (true) {
      if ((m_mbusDevTuple[m_deviceIdx].address != 0) && (m_mbusDevTuple[m_deviceIdx].timer == 0)) {
        Serial << "Issue request for device " << m_mbusDevTuple[m_deviceIdx].token << endl;
        uint8_t *sendBuffer = m_meterBusMaster->getSendBuffer();
        if (sendBuffer != 0) {
          Serial << "send buffer ready" << endl;
          m_mbusDevTuple[m_deviceIdx].timer = m_mbusDevTuple[m_deviceIdx].queryPeriod;

          sendBuffer[0] = 0x10;
          sendBuffer[1] = 0x5b;
          sendBuffer[2] = m_mbusDevTuple[m_deviceIdx].address;
          sendBuffer[3] = (uint8_t)(sendBuffer[1] + sendBuffer[2]);
          sendBuffer[4] = 0x16;
          m_meterBusMaster->sendBufferReady(5, m_mbusDevTuple[m_deviceIdx].token, this);
        } else {
          Serial << "no send buffer ready" << endl;
        }
        break;
      } else {
        //Serial << "Trying " << m_deviceIdx << ", " << m_mbusDevTuple[m_deviceIdx].token << endl;
        m_deviceIdx++;
        if (m_deviceIdx >= NUM_OF_DEVICES) {
          m_deviceIdx = 0;
        }
        break;
      }
    }
    //Serial << "while out" << endl;
  }

  //Serial << "*** d" << endl;

}



